# wadze.py
#
# Web Assembly Decoder - Zero Extras
#
# Copyright (C) 2019
# David M. Beazley (https://www.dabeaz.com)
# All Rights Reserved
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.  
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.  
# * Neither the name of the David Beazley or Dabeaz LLC may be used to
#   endorse or promote products derived from this software without
#   specific prior written permission. 
#
# This software is provided "as is."  The risks of using it are yours.

from collections import namedtuple
from itertools import islice
import struct

_typemap = {
    0x7f : 'i32', 
    0x7e : 'i64', 
    0x7d : 'f32',  
    0x7c : 'f64', 
    0x70 : 'anyfunc',
    0x60 : 'func',
    0x40 : None,
}

def parse_unsigned(stream):
    result = shift = 0
    while True:
        b = next(stream)
        result |= (b & 0x7f) << shift
        shift += 7
        if not (b & 0x80):
            break
    return result

def parse_signed(stream):
    result = shift = 0
    while True:
        b = next(stream)
        result |= (b & 0x7f) << shift
        shift += 7
        if not (b & 0x80):
            break
    if b & 0x40:
        result |= (~0 << shift)
    return result

def parse_float32(stream):
    return struct.unpack('<f', bytes(islice(stream, 4)))[0]

def parse_float64(stream):
    return struct.unpack('<d', bytes(islice(stream, 8)))[0]

def parse_vector(stream, func):
    return [ func(stream) for _ in range(parse_unsigned(stream)) ]

def parse_string(stream):
    return bytes(parse_vector(stream, next)).decode('utf-8')

FunctionType = namedtuple('FunctionType', ['params', 'returns'])

def parse_functype(stream):
    sigtype = next(stream)
    params = [ _typemap[t] for t in parse_vector(stream, next) ]
    returns = [ _typemap[t] for t in parse_vector(stream, next) ]
    return FunctionType(params, returns)

Limits = namedtuple('Limits', ['min', 'max'])

def parse_limits(stream):
    return Limits(parse_unsigned(stream), parse_unsigned(stream)) if next(stream) else (parse_unsigned(stream), None)

TableType = namedtuple('TableType', ['elemtype', 'limits'])

def parse_tabletype(stream):
    return TableType(_typemap[next(stream)], parse_limits(stream))


GlobalType = namedtuple('GlobalType', ['type', 'mut'])

def parse_globaltype(stream):
    return GlobalType(_typemap[next(stream)], next(stream))

ImportFunction = namedtuple('ImportFunction', ['module','name','typeidx'])
ImportTable = namedtuple('ImportTable', ['module','name', 'tabletype'])
ImportMemory = namedtuple('ImportMemory', ['module', 'name', 'limits'])
ImportGlobal = namedtuple('ImportGlobal', ['module', 'name', 'globaltype'])

_imports = {
    0 : (ImportFunction, parse_unsigned),
    1 : (ImportTable, parse_tabletype),
    2 : (ImportMemory, parse_limits),
    3 : (ImportGlobal, parse_globaltype),
}    

def parse_import(stream):
    module = parse_string(stream)
    name = parse_string(stream)
    cls, func = _imports[next(stream)]
    return cls(module, name, func(stream))

Global = namedtuple('Global', ['globaltype', 'expr'])

def parse_global(stream):
    return Global(parse_globaltype(stream), parse_instructions(stream))

ExportFunction = namedtuple('ExportFunction', ['name', 'ref'])
ExportTable = namedtuple('ExportTable', ['name', 'ref'])
ExportMemory = namedtuple('ExportMemory', ['name', 'ref'])
ExportGlobal = namedtuple('ExportGlobal', ['name', 'ref'])

_exports = { 0: ExportFunction, 1: ExportTable, 2: ExportMemory, 3: ExportGlobal }

def parse_export(stream):
    name = parse_string(stream)
    return _exports[next(stream)](name, parse_unsigned(stream))

Element = namedtuple('Element', ['tableidx', 'offset', 'values'])

def parse_element(stream):
    return Element(parse_unsigned(stream), parse_instructions(stream), parse_vector(stream, parse_unsigned))

def parse_locals(stream):
    return parse_unsigned(stream) * (_typemap[next(stream)],)

Code = namedtuple('Code', ['locals', 'instructions'])

def parse_rawcode(stream):
    return bytes(islice(stream, parse_unsigned(stream)))

def parse_code(raw):
    if isinstance(raw, bytes):
        code = iter(raw)
    locals = [ ]
    for loc in parse_vector(code, parse_locals):
        locals.extend(loc)
    instructions = parse_instructions(code)
    return Code(locals, instructions)

Data = namedtuple('Data', ['memidx', 'offset', 'values'])

def parse_data(stream):
    return Data(parse_unsigned(stream), parse_instructions(stream), parse_vector(stream, next))

def parse_instructions(stream):
    instructions = [ ]
    while True:
        op = next(stream)
        if op == 0x0b:
            break
        name, *funcs = _opcodes[op]
        instructions.append((name, *(func(stream) for func in funcs)))
    return instructions

def _split_else(instructions):
    if ('else',) in instructions:
        index = instructions.index(('else',))
        return (instructions[:index], instructions[index+1:])
    else:
        return (instructions, [])

_opcodes = {
    0x00 : ('unreachable', ),
    0x01 : ('nop', ),
    0x02 : ('block', lambda s: _typemap[next(s)], parse_instructions),
    0x03 : ('loop', lambda s: _typemap[next(s)], parse_instructions),
    0x04 : ('if', lambda s: _typemap[next(s)], lambda s: _split_else(parse_instructions(s))),
    0x05 : ('else',),
    0x0c : ('br', parse_unsigned),
    0x0d : ('br_if', parse_unsigned),
    0x0e : ('br_table', lambda s: parse_vector(s, parse_unsigned), parse_unsigned),
    0x0f : ('return', ),
    0x10 : ('call', parse_unsigned),
    0x11 : ('call_indirect', parse_unsigned, next),
    0x1a : ('drop', ),
    0x1b : ('select', ),
    0x20 : ('local.get', parse_unsigned),
    0x21 : ('local.set', parse_unsigned),
    0x22 : ('local.tee', parse_unsigned),
    0x23 : ('global.get', parse_unsigned),
    0x24 : ('global.set', parse_unsigned),
    0x28 : ('i32.load', parse_signed, parse_signed),
    0x29 : ('i64.load', parse_signed, parse_signed),
    0x2a : ('f32.load', parse_signed, parse_signed),
    0x2b : ('f64.load', parse_signed, parse_signed),
    0x2c : ('i32.load8_s', parse_signed, parse_signed),
    0x2d : ('i32.load8_u', parse_signed, parse_signed),
    0x2e : ('i32.load16_s', parse_signed, parse_signed),
    0x2f : ('i32.load16_u', parse_signed, parse_signed),
    0x30 : ('i64.load8_s', parse_signed, parse_signed),
    0x31 : ('i64.load8_u', parse_signed, parse_signed),
    0x32 : ('i64.load16_s', parse_signed, parse_signed),
    0x33 : ('i64.load16_u', parse_signed, parse_signed),
    0x34 : ('i64.load32_s', parse_signed, parse_signed),
    0x35 : ('i64.load32_u', parse_signed, parse_signed),
    0x36 : ('i32.store', parse_signed, parse_signed),
    0x37 : ('i64.store', parse_signed, parse_signed),
    0x38 : ('f32.store', parse_signed, parse_signed),
    0x39 : ('f64.store', parse_signed, parse_signed),
    0x3a : ('i32.store8', parse_signed, parse_signed),
    0x3b : ('i32.store16', parse_signed, parse_signed),
    0x3c : ('i64.store8', parse_signed, parse_signed),
    0x3d : ('i64.store16', parse_signed, parse_signed),
    0x3e : ('i64.store32', parse_signed, parse_signed),
    0x3f : ('memory.size', next),
    0x40 : ('memory.grow', next),
    0x41 : ('i32.const', parse_signed),
    0x42 : ('i64.const', parse_signed),
    0x43 : ('f32.const', parse_float32),
    0x44 : ('f64.const', parse_float64),
    0x45 : ('i32.eqz', ),
    0x46 : ('i32.eq', ),
    0x47 : ('i32.ne', ),
    0x48 : ('i32.lt_s', ),
    0x49 : ('i32.lt_u', ),
    0x4a : ('i32.gt_s', ),
    0x4b : ('i32.gt_u', ),
    0x4c : ('i32.le_s', ),
    0x4d : ('i32.le_u', ),
    0x4e : ('i32.ge_s', ),
    0x4f : ('i32.ge_u', ),
    0x50 : ('i64.eqz', ),
    0x51 : ('i64.eq', ),
    0x52 : ('i64.ne', ),
    0x53 : ('i64.lt_s', ),
    0x54 : ('i64.lt_u', ),
    0x55 : ('i64.gt_s', ),
    0x56 : ('i64.gt_u', ),
    0x57 : ('i64.le_s', ),
    0x58 : ('i64.le_u', ),
    0x59 : ('i64.ge_s', ),
    0x5a : ('i64.ge_u', ),
    0x5b : ('f32.eq', ),
    0x5c : ('f32.ne', ),
    0x5d : ('f32.lt', ),
    0x5e : ('f32.gt', ),
    0x5f : ('f32.le', ),
    0x60 : ('f32.ge', ),
    0x61 : ('f64.eq', ),
    0x62 : ('f64.ne', ),
    0x63 : ('f64.lt', ),
    0x64 : ('f64.gt', ),
    0x65 : ('f64.le', ),
    0x66 : ('f64.ge', ),
    0x67 : ('i32.clz', ),
    0x68 : ('i32.ctz', ),
    0x69 : ('i32.popcnt', ),
    0x6a : ('i32.add', ),
    0x6b : ('i32.sub', ),
    0x6c : ('i32.mul', ),
    0x6d : ('i32.div_s', ),
    0x6e : ('i32.div_u', ),
    0x6f : ('i32.rem_s', ),
    0x70 : ('i32.rem_u', ),
    0x71 : ('i32.and', ),
    0x72 : ('i32.or', ),
    0x73 : ('i32.xor', ),
    0x74 : ('i32.shl', ),
    0x75 : ('i32.shr_s', ),
    0x76 : ('i32.shr_u', ),
    0x77 : ('i32.rotl', ),
    0x78 : ('i32.rotr', ),
    0x79 : ('i64.clz', ),
    0x7a : ('i64.ctz', ),
    0x7b : ('i64.popcnt', ),
    0x7c : ('i64.add', ),
    0x7d : ('i64.sub', ),
    0x7e : ('i64.mul', ),
    0x7f : ('i64.div_s', ),
    0x80 : ('i64.div_u', ),
    0x81 : ('i64.rem_s', ),
    0x82 : ('i64.rem_u', ),
    0x83 : ('i64.and', ),
    0x84 : ('i64.or', ),
    0x85 : ('i64.xor', ),
    0x86 : ('i64.shl', ),
    0x87 : ('i64.shr_s', ),
    0x88 : ('i64.shr_u', ),
    0x89 : ('i64.rotl', ),
    0x8a : ('i64.rotr', ),
    0x8b : ('f32.abs', ),
    0x8c : ('f32.neg', ),
    0x8d : ('f32.ceil', ),
    0x8e : ('f32.floor', ),
    0x8f : ('f32.trunc', ),
    0x90 : ('f32.nearest', ),
    0x91 : ('f32.sqrt', ),
    0x92 : ('f32.add', ),
    0x93 : ('f32.sub', ),
    0x94 : ('f32.mul', ),
    0x95 : ('f32.div', ),
    0x96 : ('f32.min', ),
    0x97 : ('f32.max', ),
    0x98 : ('f32.copysign', ),
    0x99 : ('f64.abs', ),
    0x9a : ('f64.neg', ),
    0x9b : ('f64.ceil', ),
    0x9c : ('f64.floor', ),
    0x9d : ('f64.trunc', ),
    0x9e : ('f64.nearest', ),
    0x9f : ('f64.sqrt', ),
    0xa0 : ('f64.add', ),
    0xa1 : ('f64.sub', ),
    0xa2 : ('f64.mul', ),
    0xa3 : ('f64.div', ),
    0xa4 : ('f64.min', ),
    0xa5 : ('f64.max', ),
    0xa6 : ('f64.copysign', ),
    0xa7 : ('i32.wrap_i64', ),
    0xa8 : ('i32.trunc_f32_s', ),
    0xa9 : ('i32.trunc_f32_u', ),
    0xaa : ('i32.trunc_f64_s', ),
    0xab : ('i32.trunc_f64_u', ),
    0xac : ('i64.extend_i32_s', ),
    0xad : ('i64.extend_i32_u', ),
    0xae : ('i64.trunc_f32_s', ),
    0xaf : ('i64.trunc_f32_u', ),
    0xb0 : ('i64.trunc_f64_s', ),
    0xb1 : ('i64.trunc_f64_u', ),
    0xb2 : ('f32.convert_i32_s', ),
    0xb3 : ('f32.convert_i32_u', ),
    0xb4 : ('f32.convert_i64_s', ),
    0xb5 : ('f32.convert_i64_u', ),
    0xb6 : ('f32.demote_f64', ),
    0xb7 : ('f64.convert_i32_s', ),
    0xb8 : ('f64.convert_i32_u', ),
    0xb9 : ('f64.convert_i64_s', ),
    0xba : ('f64.convert_i64_u', ),
    0xbb : ('f64.promote_f32', ),
    0xbc : ('i32.reinterpret_f32', ),
    0xbd : ('i64.reinterpret_f64', ),
    0xbe : ('f32.reinterpret_i32', ),
    0xbf : ('f64.reinterpret_i64', ),
}

_sections = {
    1 : ('type', lambda s: parse_vector(s, parse_functype)),
    2 : ('import', lambda s: parse_vector(s, parse_import)),
    3 : ('func', lambda s: parse_vector(s, parse_unsigned)),
    4 : ('table', lambda s: parse_vector(s, parse_tabletype)),
    5 : ('memory', lambda s: parse_vector(s, parse_limits)),
    6 : ('global', lambda s: parse_vector(s, parse_global)),
    7 : ('export', lambda s: parse_vector(s, parse_export)),
    8 : ('start', parse_unsigned),
    9 : ('element', lambda s: parse_vector(s, parse_element)),
    10 : ('code', lambda s: parse_vector(s, parse_rawcode)),
    11 : ('data', lambda s: parse_vector(s, parse_data)),
}

def parse_section(stream):
    sectnum = next(stream)
    size = parse_unsigned(stream)
    if sectnum in _sections:
        sectname, parse_func = _sections[sectnum]
        return (sectname, parse_func(stream))
    else:
        return (sectnum, bytes(islice(stream, size)))

def parse_module(stream):
    '''
    Parse binary .wasm format data into a dictionary representing the
    different sections of a Wasm module. stream is either an iterator
    producing bytes or a byte string containing the raw data from a
    .wasm file.
    '''
    if isinstance(stream, bytes):
        stream = iter(stream)
    header = bytes(islice(stream, 8))
    if header[:4] != b'\x00asm':
        raise ValueError('Expected .wasm')
    sections = { }
    while True:
        try:
            sectname, items = parse_section(stream)
            sections[sectname] = items
        except StopIteration:
            break
    return sections

# Example use
if __name__ == '__main__':
    module = parse_module(open('input.wasm','rb').read())
    # If you want to parse instruction code, also include this
    module['code'] = [ parse_code(c) for c in module['code']]



        


# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from ccdexplorer_fundamentals.enums import NET
from enum import Enum
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
import os
import sys

sys.path.append(os.path.dirname("ccdexplorer_fundamentals"))
# from ccdexplorer_fundamentals.pool import ConcordiumPoolFromClient
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import CCD_PoolInfo
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient


class Mixin(_SharedConverters):
    def get_pool_info_for_pool(
        self: GRPCClient,
        pool_id: int,
        block_hash: str,
        net: Enum = NET.MAINNET,
    ) -> CCD_PoolInfo:
        prefix = ""
        result = {}
        blockHashInput = self.generate_block_hash_input_from(block_hash)
        baker_id = BakerId(value=pool_id)
        poolInfoRequest = PoolInfoRequest(baker=baker_id, block_hash=blockHashInput)

        grpc_return_value: PoolInfoResponse = self.stub_on_net(
            net, "GetPoolInfo", poolInfoRequest
        )

        for descriptor in grpc_return_value.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(
                descriptor, grpc_return_value
            )
            key_to_store = f"{prefix}{key}"
            if type(value) in self.simple_types:
                result[key_to_store] = self.convertType(value)

            elif type(value) == BakerPoolInfo:
                result[key_to_store] = self.convertBakerPoolInfo(value)

            elif type(value) == PoolCurrentPaydayInfo:
                if self.valueIsEmpty(value):
                    result[key_to_store] = None
                else:
                    result[key_to_store] = self.convertPoolCurrentPaydayInfo(value)

            elif type(value) == PoolPendingChange:
                result[key_to_store] = self.convertPoolPendingChange(value)

        return CCD_PoolInfo(**result)



from __future__ import annotations
from pydantic import BaseModel, Field, ConfigDict
import datetime as dt
from typing import Union, Optional
from enum import Enum
import io
import base58
from ccdexplorer_fundamentals.GRPCClient import GRPCClient
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *  # noqa: F403
from ccdexplorer_fundamentals.mongodb import Collections
from pymongo.collection import Collection
from pymongo import ReplaceOne
from ccdexplorer_fundamentals.enums import NET
from rich.console import Console

# import math
import leb128

console = Console()

LEN_ACCOUNT_ADDRESS = 50


# Metadata classes
class TokenAttribute(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    type: Optional[str] = None
    name: Optional[str] = None
    value: Optional[str] = None


class TokenURLJSON(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    url: Optional[str] = None
    hash: Optional[str] = None


class TokenMetaData(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    name: Optional[str] = None
    symbol: Optional[str] = None
    unique: Optional[bool] = None
    decimals: Optional[int] = None
    description: Optional[str] = None
    thumbnail: Optional[TokenURLJSON] = None
    display: Optional[TokenURLJSON] = None
    artifact: Optional[TokenURLJSON] = None
    assets: Optional[list[TokenMetaData]] = None
    attributes: Optional[list[TokenAttribute]] = None
    localization: Optional[dict[str, TokenURLJSON]] = None


# class MongoTypeTokenHolder(BaseModel):
#     account_address: CCD_AccountAddress
#     token_amount: str


class MongoTypeTokenForAddress(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    token_address: str
    contract: str
    token_id: str
    token_amount: str


class MongoTypeTokenLink(BaseModel):
    id: str = Field(..., alias="_id")
    account_address: Optional[str] = None
    account_address_canonical: Optional[str] = None
    token_holding: Optional[MongoTypeTokenForAddress] = None


class MongoTypeTokenHolderAddress(BaseModel):
    id: str = Field(..., alias="_id")
    account_address_canonical: Optional[str] = None
    tokens: dict[str, MongoTypeTokenForAddress]


class MongoTypeLoggedEvent(BaseModel):
    id: str = Field(..., alias="_id")
    logged_event: str
    result: dict
    tag: int
    event_type: str
    block_height: int
    slot_time: Optional[dt.datetime] = None
    tx_index: int  #####################################################################
    ordering: int
    tx_hash: str
    token_address: str
    contract: str
    date: Optional[str] = None


class MongoTypeTokensTag(BaseModel):
    id: str = Field(..., alias="_id")
    contracts: list[str]
    tag_template: Optional[bool] = None
    single_use_contract: Optional[bool] = None
    logo_url: Optional[str] = None
    decimals: Optional[int] = None
    exchange_rate: Optional[float] = None
    get_price_from: Optional[str] = None
    logged_events_count: Optional[int] = None
    owner: Optional[str] = None
    module_name: Optional[str] = None
    token_type: Optional[str] = None
    display_name: Optional[str] = None
    tvl_for_token_in_usd: Optional[float] = None
    token_tag_id: Optional[str] = None


class FailedAttempt(BaseModel):
    attempts: int
    do_not_try_before: dt.datetime
    last_error: str


class MongoTypeTokenAddress(BaseModel):
    id: str = Field(..., alias="_id")
    contract: str
    token_id: str
    token_amount: Optional[str] = None
    metadata_url: Optional[str] = None
    last_height_processed: int
    token_holders: Optional[dict[str, str]] = None
    tag_information: Optional[MongoTypeTokensTag] = None
    exchange_rate: Optional[float] = None
    domain_name: Optional[str] = None
    token_metadata: Optional[TokenMetaData] = None
    failed_attempt: Optional[FailedAttempt] = None
    hidden: Optional[bool] = None


# CIS
class StandardIdentifiers(Enum):
    CIS_0 = "CIS-0"
    CIS_1 = "CIS-1"
    CIS_2 = "CIS-2"
    CIS_3 = "CIS-3"
    CIS_4 = "CIS-4"
    CIS_5 = "CIS-5"
    CIS_6 = "CIS-6"


class LoggedEvents(Enum):
    transfer_event = 255
    mint_event = 254
    burn_event = 253
    operator_event = 252
    metadata_event = 251
    nonce_event = 250
    register_credential_event = 249
    revoke_credential_event = 248
    issuer_metadata_event = 247
    credential_metadata_event = 246
    credential_schemaref_event = 245
    recovation_key_event = 244
    item_created_event = 237
    item_status_changed = 236


# CIS-2 Logged Event Types


class transferEvent(BaseModel):
    tag: int
    token_id: Optional[str] = None
    token_amount: Optional[int] = None
    from_address: Optional[str] = None
    to_address: Optional[str] = None


class mintEvent(BaseModel):
    tag: int
    token_id: Optional[str] = None
    token_amount: Optional[int] = None
    to_address: Optional[str] = None


class burnEvent(BaseModel):
    tag: int
    token_id: Optional[str] = None
    token_amount: Optional[int] = None
    from_address: Optional[str] = None


class updateOperatorEvent(BaseModel):
    tag: int
    operator_update: Optional[str] = None
    owner: Optional[str] = None
    operator: Optional[str] = None


class SchemaRef(BaseModel):
    url: str
    checksum: Optional[str] = None


class registerCredentialEvent(BaseModel):
    tag: int
    credential_id: Optional[str] = None
    schema_ref: Optional[SchemaRef] = None
    credential_type: Optional[str] = None


class revokeCredentialEvent(BaseModel):
    tag: int
    credential_id: Optional[str] = None
    revoker: Optional[str] = None
    reason: Optional[str] = None


class issuerMetadataEvent(BaseModel):
    tag: int
    metadata: MetadataUrl


class credentialMetadataEvent(BaseModel):
    tag: int
    id: str  # credentialHolderId
    metadata: MetadataUrl


class credentialSchemaRefEvent(BaseModel):
    tag: int
    type: Optional[str] = None
    schema_ref: Optional[str] = None


class revocationKeyEvent(BaseModel):
    tag: int
    action: Optional[str] = None


class MetadataUrl(BaseModel):
    url: str
    checksum: Optional[str] = None


class tokenMetadataEvent(BaseModel):
    tag: int
    token_id: str
    metadata: MetadataUrl


class itemCreatedEvent(BaseModel):
    tag: int
    item_id: str
    metadata: MetadataUrl
    initial_status: str


class itemStatusChangedEvent(BaseModel):
    tag: int
    item_id: str
    new_status: str
    additional_data: str


class nonceEvent(BaseModel):
    tag: int
    nonce: Optional[int] = None
    sponsoree: Optional[str] = None


class CIS:
    def __init__(
        self,
        grpcclient: GRPCClient = None,
        instance_index=None,
        instance_subindex=None,
        entrypoint=None,
        net: NET.MAINNET = None,
    ):
        self.grpcclient = grpcclient
        self.instance_index = instance_index
        self.instance_subindex = instance_subindex
        self.entrypoint = entrypoint
        self.net = net

    ###############

    def execute_save(self, collection: Collection, replacement, _id: str):
        repl_dict = replacement.dict()
        if "id" in repl_dict:
            del repl_dict["id"]

        # sort tokens and token_holders
        if "tokens" in repl_dict:
            sorted_tokens = list(repl_dict["tokens"].keys())
            sorted_tokens.sort()
            tokens_sorted = {i: repl_dict["tokens"][i] for i in sorted_tokens}
            repl_dict["tokens"] = tokens_sorted

        if "token_holders" in repl_dict:
            sorted_holders = list(repl_dict["token_holders"].keys())
            sorted_holders.sort()
            token_holders_sorted = {
                i: repl_dict["token_holders"][i] for i in sorted_holders
            }
            repl_dict["token_holders"] = token_holders_sorted

        _ = collection.bulk_write(
            [
                ReplaceOne(
                    {"_id": _id},
                    replacement=repl_dict,
                    upsert=True,
                )
            ]
        )

    def restore_state_for_token_address(
        self,
        db_to_use: dict[Collections, Collection],
        token_address: str,
    ):
        d: dict = db_to_use[Collections.tokens_token_addresses].find_one(
            {"_id": token_address}
        )

        d.update(
            {
                "token_amount": str(int(0)),  # mongo limitation on int size
                "token_holders": {},  # {CCD_AccountAddress, str(token_amount)}
                "last_height_processed": 0,
            }
        )

        d = MongoTypeTokenAddress(**d)
        self.execute_save(
            db_to_use[Collections.tokens_token_addresses], d, token_address
        )

    def copy_token_holders_state_to_address_and_save(
        self,
        db_to_use: dict[Collections, Collection],
        token_address_info: MongoTypeTokenAddress,
        address: str,
    ):
        token_address = token_address_info.id
        d = db_to_use[Collections.tokens_accounts].find_one({"_id": address})
        # if this account doesn't have tokens, create empty dict.
        if not d:
            d = MongoTypeTokenHolderAddress(
                **{
                    "_id": address,
                    "tokens": {},
                }
            )  # keyed on token_address
        else:
            d = MongoTypeTokenHolderAddress(**d)

        token_to_save = MongoTypeTokenForAddress(
            **{
                "token_address": token_address,
                "contract": token_address_info.contract,
                "token_id": token_address_info.token_id,
                "token_amount": str(token_address_info.token_holders.get(address, 0)),
            }
        )

        d.tokens[token_address] = token_to_save

        if token_to_save.token_amount == str(0):
            del d.tokens[token_address]

        self.execute_save(db_to_use[Collections.tokens_accounts], d, address)

    def save_mint(
        self,
        db_to_use: dict[Collections, Collection],
        instance_address: str,
        result: mintEvent,
        height: int,
    ):
        token_address = f"{instance_address}-{result.token_id}"
        d = db_to_use[Collections.tokens_token_addresses].find_one(
            {"_id": token_address}
        )
        if not d:
            d = MongoTypeTokenAddress(
                **{
                    "_id": token_address,
                    "contract": instance_address,
                    "token_id": result.token_id,
                    "token_amount": str(int(0)),  # mongo limitation on int size
                    "token_holders": {},  # {CCD_AccountAddress, str(token_amount)}
                    "last_height_processed": height,
                }
            )
        else:
            d = MongoTypeTokenAddress(**d)

        token_holders: dict[CCD_AccountAddress, str] = d.token_holders  # noqa: F405
        token_holders[result.to_address] = str(
            int(token_holders.get(result.to_address, "0")) + result.token_amount
        )
        d.token_amount = str((int(d.token_amount) + result.token_amount))
        d.token_holders = token_holders

        self.execute_save(
            db_to_use[Collections.tokens_token_addresses], d, token_address
        )
        self.copy_token_holders_state_to_address_and_save(
            db_to_use, d, result.to_address
        )

    def save_metadata(
        self,
        db_to_use: dict[Collections, Collection],
        instance_address: str,
        result: tokenMetadataEvent,
        height: int,
    ):
        token_address = f"{instance_address}-{result.token_id}"
        d = db_to_use[Collections.tokens_token_addresses].find_one(
            {"_id": token_address}
        )
        if not d:
            d = MongoTypeTokenAddress(
                **{
                    "_id": token_address,
                    "contract": instance_address,
                    "token_id": result.token_id,
                    "last_height_processed": height,
                }
            )
        else:
            d = MongoTypeTokenAddress(**d)

        d.metadata_url = result.metadata.url

        self.execute_save(
            db_to_use[Collections.tokens_token_addresses], d, token_address
        )

    def save_operator(
        self,
        db_to_use: dict[Collections, Collection],
        instance_address: str,
        result: tokenMetadataEvent,
        height: int,
    ):
        token_address = f"{instance_address}-{result.token_id}"
        d = db_to_use[Collections.tokens_token_addresses].find_one(
            {"_id": token_address}
        )
        if not d:
            d = MongoTypeTokenAddress(
                **{
                    "_id": token_address,
                    "contract": instance_address,
                    "token_id": result.token_id,
                    "last_height_processed": height,
                }
            )
        else:
            d = MongoTypeTokenAddress(**d)

        d.metadata_url = result.metadata.url

        self.execute_save(
            db_to_use[Collections.tokens_token_addresses], d, token_address
        )

    def save_transfer(
        self,
        db_to_use: dict[Collections, Collection],
        instance_address: str,
        result: transferEvent,
        height: int,
    ):
        token_address = f"{instance_address}-{result.token_id}"
        d = db_to_use[Collections.tokens_token_addresses].find_one(
            {"_id": token_address}
        )
        if not d:
            return None

        d = MongoTypeTokenAddress(**d)

        try:
            token_holders: dict[CCD_AccountAddress, str] = d.token_holders  # noqa: F405
        except:  # noqa: E722
            console.log(
                f"{result.tag}: {token_address} | {d} has no field token_holders?"
            )
            Exception(
                console.log(
                    f"{result.tag}: {token_address} | {d} has no field token_holders?"
                )
            )

        token_holders[result.to_address] = str(
            int(token_holders.get(result.to_address, "0")) + result.token_amount
        )
        try:
            token_holders[result.from_address] = str(
                int(token_holders.get(result.from_address, None)) - result.token_amount
            )
            if int(token_holders[result.from_address]) >= 0:
                d.token_holders = token_holders
                d.last_height_processed = height
                self.execute_save(
                    db_to_use[Collections.tokens_token_addresses], d, token_address
                )

                self.copy_token_holders_state_to_address_and_save(
                    db_to_use, d, result.from_address
                )
                self.copy_token_holders_state_to_address_and_save(
                    db_to_use, d, result.to_address
                )

        except:  # noqa: E722
            if result.token_amount > 0:
                console.log(
                    f"""{result.tag}: {result.from_address} is not listed 
                    as token holder for {token_address}?"""
                )

    def save_burn(
        self,
        db_to_use: dict[Collections, Collection],
        instance_address: str,
        result: burnEvent,
        height: int,
    ):
        token_address = f"{instance_address}-{result.token_id}"

        d = MongoTypeTokenAddress(
            **db_to_use[Collections.tokens_token_addresses].find_one(
                {"_id": token_address}
            )
        )

        token_holders: dict[CCD_AccountAddress, str] = d.token_holders  # noqa: F405
        try:
            token_holders[result.from_address] = str(
                int(token_holders.get(result.from_address, "0")) - result.token_amount
            )
            if token_holders[result.from_address] == str(0):
                del token_holders[result.from_address]

            d.token_amount = str((int(d.token_amount) - result.token_amount))
            d.token_holders = token_holders
            d.last_height_processed = height

            if int(d.token_amount) >= 0:
                self.execute_save(
                    db_to_use[Collections.tokens_token_addresses], d, token_address
                )
                self.copy_token_holders_state_to_address_and_save(
                    db_to_use, d, result.from_address
                )

        except:  # noqa: E722
            console.log(
                f"""{result.tag}: {result.from_address} is not listed as 
                token holder for {token_address}?"""
            )
            # exit

    def formulate_logged_event(
        self,
        slot_time: dt.datetime,
        tag_: int,
        result: Union[
            mintEvent, burnEvent, transferEvent, updateOperatorEvent, tokenMetadataEvent
        ],
        instance_address: str,
        event: str,
        height: int,
        tx_hash: str,
        tx_index: int,
        ordering: int,
        _id_postfix: str,
    ) -> Union[ReplaceOne, None]:
        if tag_ in [255, 254, 253, 252, 251, 250]:
            if tag_ == 252:
                token_address = f"{instance_address}-operator"
            elif tag_ == 250:
                token_address = f"{instance_address}-nonce"
            else:
                token_address = f"{instance_address}-{result.token_id}"
            _id = f"{height}-{token_address}-{event}-{_id_postfix}"
            if result:
                result_dict = result.model_dump()
            else:
                result_dict = {}
            if "token_amount" in result_dict:
                result_dict["token_amount"] = str(result_dict["token_amount"])

            d = {
                "_id": _id,
                "logged_event": event,
                "result": result_dict,
                "tag": tag_,
                "event_type": LoggedEvents(tag_).name,
                "block_height": height,
                "tx_hash": tx_hash,
                "tx_index": tx_index,
                "ordering": ordering,
                "token_address": token_address,
                "contract": instance_address,
                "date": f"{slot_time:%Y-%m-%d}",
            }
            return (
                MongoTypeLoggedEvent(**d),
                ReplaceOne(
                    {"_id": _id},
                    replacement=d,
                    upsert=True,
                ),
            )

        else:
            return (None, None)

    # not used
    def execute_logged_event(
        self,
        db_to_use,
        tag_: int,
        result: Union[mintEvent, burnEvent, transferEvent, tokenMetadataEvent],
        instance_address: str,
        height: int,
    ):
        if tag_ == 255:
            self.save_transfer(db_to_use, instance_address, result, height)
        elif tag_ == 254:
            self.save_mint(db_to_use, instance_address, result, height)
        elif tag_ == 253:
            self.save_burn(db_to_use, instance_address, result, height)
        elif tag_ == 252:
            pass
            # we only save the logged event, but to not process this in
            # token_address or accounts.
            # save_operator(db_to_use, instance_address, result, height)
        elif tag_ == 251:
            self.save_metadata(db_to_use, instance_address, result, height)
        elif tag_ == 250:
            pass  # nonceEvent

    def process_event(
        self,
        slot_time: dt.datetime,
        instance_address: str,
        event: str,
        height: int,
        tx_hash: str,
        tx_index: int,
        ordering: int,
        _id_postfix: str,
    ):
        tag_, result = self.process_log_events(event)
        logged_event = None
        logged_event_for_queue = None
        token_address = None
        if result:
            # if tag_ in [255, 254, 253, 252, 251, 250]:
            if tag_ in [255, 254, 253, 251]:
                token_address = f"{instance_address}-{result.token_id}"

                (logged_event, logged_event_for_queue) = self.formulate_logged_event(
                    slot_time,
                    tag_,
                    result,
                    instance_address,
                    event,
                    height,
                    tx_hash,
                    tx_index,
                    ordering,
                    _id_postfix,
                )

        return tag_, logged_event, logged_event_for_queue, token_address

    ###############

    def standard_identifier(self, identifier: StandardIdentifiers) -> bytes:
        si = io.BytesIO()
        # write the length of ASCII characters for the identifier
        number = len(identifier.value)
        byte_array = number.to_bytes(1, "little")
        si.write(byte_array)
        # write the identifier
        si.write(bytes(identifier.value, encoding="ASCII"))
        # convert to bytes
        return si.getvalue()

    def supports_parameter(self, standard_identifier: StandardIdentifiers) -> bytes:
        sp = io.BytesIO()
        # write the number of standardIdentifiers present
        number = 1
        byte_array = number.to_bytes(2, "little")
        sp.write(byte_array)
        # write the standardIdentifier
        sp.write(self.standard_identifier(standard_identifier))
        # convert to bytes
        return sp.getvalue()

    def support_result(self, bs: io.BytesIO):
        t = int.from_bytes(bs.read(2), byteorder="little")
        if t == 0:
            return t, "Standard is not supported"
        elif t == 1:
            return t, "Standard is supported by this contract"
        elif t == 2:
            contracts = []
            n = int.from_bytes(bs.read(1), byteorder="little")
            for _ in range(n):
                contracts.append(self.contract_address(bs))
                return (
                    t,
                    "Standard is supported by using one of these contract addresses: "
                    + [x for x in contracts],
                )

    def supports_response(self, res: bytes):
        bs = io.BytesIO(bytes.fromhex(res.decode()))
        if bs.getbuffer().nbytes > 0:
            n = int.from_bytes(bs.read(2), byteorder="little")
            responses = []
            for _ in range(n):
                responses.append(self.support_result(bs))
            if len(responses) > 0:
                return responses[0]
            else:
                return False, "Lookup Failure"
        else:
            return False, "Lookup Failure"

    def supports_standard(self, standard_identifier: StandardIdentifiers) -> bool:
        parameter_bytes = self.supports_parameter(standard_identifier)

        ii = self.grpcclient.invoke_instance(
            "last_final",
            self.instance_index,
            self.instance_subindex,
            self.entrypoint,
            parameter_bytes,
            self.net,
        )

        res = ii.success.return_value
        support_result, support_result_text = self.supports_response(res)

        return support_result == 1

    def supports_standards(
        self, standard_identifiers: list[StandardIdentifiers]
    ) -> bool:
        support = False
        for si in standard_identifiers:
            parameter_bytes = self.supports_parameter(si)

            ii = self.grpcclient.invoke_instance(
                "last_final",
                self.instance_index,
                self.instance_subindex,
                self.entrypoint,
                parameter_bytes,
                self.net,
            )

            res = ii.success.return_value
            support_result, support_result_text = self.supports_response(res)

            support = support_result == 1
        return support

    def balanceOf(self, block_hash: str, tokenID: str, account_id: str):
        parameter_bytes = self.balanceOfParameter(tokenID, account_id)

        ii = self.grpcclient.invoke_instance(
            block_hash,
            self.instance_index,
            self.instance_subindex,
            self.entrypoint,
            parameter_bytes,
            self.net,
        )

        res = ii.success.return_value
        support_result = self.balanceOfResponse(res)

        return support_result, ii

    def account_address(self, bs: io.BytesIO):
        addr = bs.read(32)
        return base58.b58encode_check(b"\x01" + addr).decode()

    def contract_address(self, bs: io.BytesIO):
        return int.from_bytes(bs.read(8), byteorder="little"), int.from_bytes(
            bs.read(8), byteorder="little"
        )

    def address(self, bs: io.BytesIO):
        t = int.from_bytes(bs.read(1), byteorder="little")
        if t == 0:
            return self.account_address(bs)
        elif t == 1:
            return self.contract_address(bs)
        else:
            raise Exception("invalid type")

    def receiver(self, bs: io.BytesIO):
        t = int.from_bytes(bs.read(1), byteorder="little")
        if t == 0:
            return self.account_address(bs)
        elif t == 1:
            return self.contract_address(bs), self.receiveHookName(bs)
        else:
            raise Exception("invalid type")

    def url(self, n: int, bs: io.BytesIO):
        data = bs.read(n)
        return data

    def metadataChecksum(self, bs: io.BytesIO):
        t = int.from_bytes(bs.read(1), byteorder="little")
        if t == 0:
            return None
        elif t == 1:
            try:
                return bs.read(32).hex()
            except:  # noqa: E722
                return -1
        else:
            # should not happen
            return -2

    def metadataUrl(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(2), byteorder="little")
        url = bs.read(n).decode()
        checksum = self.metadataChecksum(bs)
        return MetadataUrl(**{"url": url, "checksum": checksum})

    def schema_ref(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(2), byteorder="little")
        url = bs.read(n).decode()
        return SchemaRef(**{"url": url, "checksum": None})

    def receiveHookName(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(2), byteorder="little")
        name = bs.read(n)
        return bytes.decode(name, "UTF-8")

    def additionalData(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(2), byteorder="little")
        return bytes.hex(bs.read(n))

    def tokenID(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        return bytes.hex(bs.read(n))

    def balanceOfQuery(self, tokenID: str, address: str):
        sp = io.BytesIO()

        tokenID = self.generate_tokenID(tokenID)
        address = self.generate_address(address)
        sp.write(tokenID)
        sp.write(address)
        return sp.getvalue()

    def balanceOfParameter(self, tokenID: str, address: str):
        sp = io.BytesIO()

        sp.write(int(1).to_bytes(2, "little"))
        sp.write(self.balanceOfQuery(tokenID, address))
        # convert to bytes
        return sp.getvalue()

    def balanceOfResponse(self, res: bytes):
        bs = io.BytesIO(bytes.fromhex(res.decode()))
        n = int.from_bytes(bs.read(2), byteorder="little")
        print(n)
        result = self.token_amount(bs)

        return result

    def generate_tokenID(self, tokenID: str):
        sp = io.BytesIO()
        tokenID_in_bytes = bytes.fromhex(tokenID)
        sp.write(int(len(tokenID_in_bytes)).to_bytes(1, "little"))
        sp.write(tokenID_in_bytes)
        return sp.getvalue()

    def generate_account_address(self, address: str):
        return bytearray(base58.b58decode_check(address)[1:])

    def generate_contract_address(self, address: str):
        # TODO
        sp = io.BytesIO()

        address_in_bytes = bytes.fromhex(address.encode("utf-8"))

        sp.write(address_in_bytes)
        return sp.getvalue()

    def generate_address(self, address: str):
        sp = io.BytesIO()

        if len(address) == 50:
            sp.write(int(0).to_bytes(1, "little"))
            sp.write(self.generate_account_address(address))
        else:
            sp.write(int(1).to_bytes(1, "little"))
            sp.write(self.generate_contract_address(address))

        return sp.getvalue()

    def invoke_token_metadataUrl(self, tokenID: str) -> bool:
        parameter_bytes = self.tokenMetadataParameter(tokenID)

        ii = self.grpcclient.invoke_instance(
            "last_final",
            self.instance_index,
            self.instance_subindex,
            self.entrypoint,
            parameter_bytes,
            self.net,
        )

        res = ii.success.return_value
        return self.tokenMetadataResultParameter(res)

    def viewOwnerHistoryRequest(self, tokenID: str):
        return self.generate_tokenID(tokenID)

    def viewOwnerHistoryResponse(self, res: bytes):
        bs = io.BytesIO(bytes.fromhex(res.decode()))
        n = int.from_bytes(bs.read(1), byteorder="little")
        _ = bs.read(3)  # own_str
        results = []
        for _ in range(0, n):
            results.append(self.address(bs))

        return results

    def tokenMetadataParameter(self, tokenID: str):
        sp = io.BytesIO()
        sp.write(int(1).to_bytes(2, "little"))
        sp.write(self.generate_tokenID(tokenID))
        return sp.getvalue()

    def metadata_result(self, bs: bytes):
        n = int(bs[:2].decode("ASCII"))
        bs = io.BytesIO(bs)
        bs.read(2)
        url = self.url(n, bs)
        return url

    def metadata_response(self, bs: bytes):
        # bs: io.BytesIO = io.BytesIO(bs)
        if len(bs) > 0:
            n = int(bs[:2].decode("ASCII"))
            # n = int.from_bytes(bs.read(2), byteorder="big")
            responses = []
            for _ in range(n):
                responses.append(self.metadata_result(bs))
            return responses[0]
        else:
            return False, "Lookup Failure"

    def tokenMetadataResultParameter(self, res: bytes):
        bs = io.BytesIO(bytes.fromhex(res.decode()))
        n = int.from_bytes(bs.read(2), byteorder="little")
        results = []
        for _ in range(0, n):
            results.append(self.metadataUrl(bs))

        return results

    def operator_update(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        if n == 0:
            return "Remove operator"
        elif n == 1:
            return "Add operator"

    def token_id(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        return bytes.hex(bs.read(n))

    def item_id(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        return bytes.hex(bs.read(n))

    def status(self, bs: io.BytesIO):
        return bytes.hex(bs.read(1))

    def nonce(self, bs: io.BytesIO):
        return bytes.hex(bs.read(8))

    def token_amount(self, bs: io.BytesIO):
        return leb128.u.decode_reader(bs)[0]

    def credential_id(self, bs: io.BytesIO):
        return bytes.hex(bs.read(32))

    def credential_type(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        return bytes.hex(bs.read(n))

    def reason_string(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        return bytes.hex(bs.read(n))

    def revoker(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        if n == 0:
            return "Issuer"
        elif n == 1:
            return "Holder"
        elif n == 2:
            key_ = self.credential_id(bs)
            return f"Other ({key_})"

    def optional_reason(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        if n == 0:
            return None
        elif n == 2:
            reason_string_ = self.reason_string(bs)
            return reason_string_

    def revocation_key_action(self, bs: io.BytesIO):
        n = int.from_bytes(bs.read(1), byteorder="little")
        if n == 0:
            return "Register"
        elif n == 1:
            return "Remove"

    def transferEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        token_id_ = self.token_id(bs)
        amount_ = self.token_amount(bs)

        from_ = self.address(bs)
        if type(from_) is not tuple:
            # it's an account address
            if len(from_) != LEN_ACCOUNT_ADDRESS:
                return None

        if type(from_) == tuple:
            from_ = f"<{from_[0]},{from_[1]}>"
        to_ = self.address(bs)
        if type(to_) is not tuple:
            # it's an account address
            if len(to_) != LEN_ACCOUNT_ADDRESS:
                return None

        if type(to_) == tuple:
            to_ = f"<{to_[0]},{to_[1]}>"

        return transferEvent(
            **{
                "tag": tag_,
                "token_id": token_id_,
                "token_amount": amount_,
                "from_address": from_,
                "to_address": to_,
            }
        )

    def updateOperatorEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        # token_id_ = self.token_id(bs)
        update_ = self.operator_update(bs)

        owner_ = self.address(bs)
        if type(owner_) is not tuple:
            # it's an account address
            if len(owner_) != LEN_ACCOUNT_ADDRESS:
                return None

        if type(owner_) == tuple:
            owner_ = f"<{owner_[0]},{owner_[1]}>"
        operator_ = self.address(bs)
        if type(operator_) is not tuple:
            # it's an account address
            if len(operator_) != LEN_ACCOUNT_ADDRESS:
                return None

        if type(operator_) == tuple:
            operator_ = f"<{operator_[0]},{operator_[1]}>"

        return updateOperatorEvent(
            **{
                "tag": tag_,
                "operator_update": update_,
                "owner": owner_,
                "operator": operator_,
            }
        )

    def mintEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        token_id_ = self.token_id(bs)
        amount_ = self.token_amount(bs)
        to_ = self.address(bs)
        if type(to_) is not tuple:
            # it's an account address
            if len(to_) != LEN_ACCOUNT_ADDRESS:
                return None

        if type(to_) == tuple:
            to_ = f"<{to_[0]},{to_[1]}>"

        return mintEvent(
            **{
                "tag": tag_,
                "token_id": token_id_,
                "token_amount": amount_,
                "to_address": to_,
            }
        )

    def burnEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        token_id_ = self.token_id(bs)
        amount_ = self.token_amount(bs)
        from_ = self.address(bs)
        if type(from_) is not tuple:
            # it's an account address
            if len(from_) != LEN_ACCOUNT_ADDRESS:
                return None

        if type(from_) == tuple:
            from_ = f"<{from_[0]},{from_[1]}>"

        return burnEvent(
            **{
                "tag": tag_,
                "token_id": token_id_,
                "token_amount": amount_,
                "from_address": from_,
            }
        )

    def tokenMetaDataEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")

        token_id_ = self.token_id(bs)
        metadata_ = self.metadataUrl(bs)

        return tokenMetadataEvent(
            **{
                "tag": tag_,
                "token_id": token_id_,
                "metadata": metadata_,
            }
        )

    def nonceEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")

        nonce_ = self.nonce(bs)
        sponsoree_ = self.account_address(bs)

        return nonceEvent(
            **{
                "tag": tag_,
                "nonce": nonce_,
                "sponsoree": sponsoree_,
            }
        )

    def registerCredentialEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        credential_id_ = self.credential_id(bs)
        schema_ref_ = self.schema_ref(bs)
        credential_type_ = self.credential_type(bs)
        return registerCredentialEvent(
            **{
                "tag": tag_,
                "credential_id": credential_id_,
                "schema_ref": schema_ref_,
                "credential_type": credential_type_,
            }
        )

    def revokeCredentialEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        credential_id_ = self.credential_id(bs)
        revoker_ = self.revoker(bs)
        reason_ = self.optional_reason(bs)
        return revokeCredentialEvent(
            **{
                "tag": tag_,
                "credential_id": credential_id_,
                "revoker": revoker_,
                "reason": reason_,
            }
        )

    def issuerMetaDataEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        metadata_ = self.metadataUrl(bs)

        return issuerMetadataEvent(
            **{
                "tag": tag_,
                "metadata": metadata_,
            }
        )

    def credentialMetaDataEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        credential_id_ = self.credential_id(bs)
        metadata_ = self.metadataUrl(bs)

        return credentialMetadataEvent(
            **{
                "tag": tag_,
                "id": credential_id_,
                "metadata": metadata_,
            }
        )

    def credentialSchemaRefEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        schema_ref_ = self.schema_ref(bs)
        credential_type_ = self.credential_type(bs)
        return credentialSchemaRefEvent(
            **{
                "tag": tag_,
                "type": credential_type_,
                "schema_ref": schema_ref_,
            }
        )

    def revocationKeyEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        action = self.revocation_key_action(bs)

        return revocationKeyEvent(
            **{
                "tag": tag_,
                "action": action,
            }
        )

    def ItemCreatedEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")

        item_id_ = self.item_id(bs)
        metadata_ = self.metadataUrl(bs)
        initial_status_ = self.status(bs)

        return itemCreatedEvent(
            **{
                "tag": tag_,
                "item_id": item_id_,
                "metadata": metadata_,
                "initial_status": initial_status_,
            }
        )

    def ItemStatusChangedEvent(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")

        item_id_ = self.item_id(bs)
        new_status_ = self.status(bs)
        additional_data_ = self.additionalData(bs)

        return itemStatusChangedEvent(
            **{
                "tag": tag_,
                "item_id": item_id_,
                "new_status": new_status_,
                "additional_data": additional_data_,
            }
        )

    def process_tnt_log_event(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        if tag_ == 237:
            try:
                event = self.ItemCreatedEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 236:
            try:
                event = self.ItemStatusChangedEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        else:
            return tag_, f"Custom even with tag={tag_}."

    def process_log_events(self, hexParameter: str):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        if tag_ == 255:
            try:
                event = self.transferEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 254:
            try:
                event = self.mintEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 253:
            try:
                event = self.burnEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 252:
            try:
                event = self.updateOperatorEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 251:
            try:
                event = self.tokenMetaDataEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 250:
            try:
                event = self.nonceEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 249:
            try:
                event = self.registerCredentialEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 248:
            try:
                event = self.revokeCredentialEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 247:
            try:
                event = self.issuerMetaDataEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 246:
            try:
                event = self.credentialMetaDataEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 245:
            try:
                event = self.credentialSchemaRefEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        elif tag_ == 244:
            try:
                event = self.revocationKeyEvent(hexParameter)
                return tag_, event
            except:  # noqa: E722
                return tag_, None
        else:
            return tag_, f"Custom even with tag={tag_}."



# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from ccdexplorer_fundamentals.enums import NET
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
import os
import sys

sys.path.append(os.path.dirname("ccdexplorer_fundamentals"))
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *


class Mixin(_SharedConverters):
    def convertElectionInfoBaker(self, message) -> list[CCD_ElectionInfo_Baker]:
        bakers = []

        for list_entry in message:
            bakers.append(
                CCD_ElectionInfo_Baker(**self.convertTypeWithSingleValues(list_entry))
            )

        return bakers

    def get_election_info(
        self: GRPCClient,
        block_hash: str,
        net: Enum = NET.MAINNET,
    ) -> CCD_ElectionInfo:
        result = {}
        blockHashInput = self.generate_block_hash_input_from(block_hash)

        grpc_return_value: ElectionInfo = self.stub_on_net(
            net, "GetElectionInfo", blockHashInput
        )

        for descriptor in grpc_return_value.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(
                descriptor, grpc_return_value
            )

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif key == "baker_election_info":
                result[key] = self.convertElectionInfoBaker(value)

        return CCD_ElectionInfo(**result)



import requests
from rich.console import Console
console = Console()
import sys

import datetime as dt

class Mixin:
    def ql_get_next_payday(self):
            query = """
                    query { paydayStatus { 
                    nextPaydayTime 
                    paydaySummaries (first: 1) {
                        nodes {
                            block {
                            blockSlotTime
                            blockHeight
                            blockHash
                            }
                        }
                        }
                    } }
            """
            try:
                r = requests.post(self.graphql_url, json={'query': query})
                if r.status_code == 200:
                    return r.json()['data']['paydayStatus']['nextPaydayTime'], \
                           r.json()['data']['paydayStatus']['paydaySummaries']['nodes'][0]['block']
                    
                else:
                    console.log(r.text)
            except Exception as e:
                console.log(query, e)
                return None

    def ql_request_paydays(self, before: str = None, after: str = None):
            row_count = (dt.datetime.now() - dt.datetime(2022,6,24,8,0,0)).days + 1
            query = "query {"
            query += f'paydayStatus {{ nextPaydayTime '
            
            if before == 'first':
                accounts_str  =  f'paydaySummaries (first: {self.nodes_request_limit}) '
            
            elif before != 'null':
                accounts_str  =  f'paydaySummaries (last: {self.nodes_request_limit}, before: "{before}") '
            elif 'last' in after:
                accounts_str  =  f'paydaySummaries ({after}) ' 
            
            elif after != 'null':
                accounts_str  =  f'paydaySummaries (first: {self.nodes_request_limit}, after: "{after}") ' 
            else:
                accounts_str = f'paydaySummaries (first: {self.nodes_request_limit})'
            query += accounts_str + '{'
            query += self.pageInfo()
            query += """
                        nodes {
                            block {
                                blockHeight
                                blockSlotTime
                                blockHash
                            }
                        }
            """
            query += ' } } }'

            try:
                r = requests.post(self.graphql_url, json={'query': query})
                if r.status_code == 200:
                    return  r.json()['data']['paydayStatus']['paydaySummaries']['nodes'], \
                            r.json()['data']['paydayStatus']['paydaySummaries']['pageInfo'],\
                            r.json()['data']['paydayStatus']['nextPaydayTime'],\
                            row_count
                        
                else:
                    console.log(r.text)
                    sys.exit(1)
            except Exception as e:
                console.log(query, e)
                sys.exit(1)
                return None

    def ql_get_all_paydays(self):
        done = False
        after = 'null'
        paydays = []
       
        while not done:
            this_batch, pageInfo = self.ql_request_paydays(before='null', after=after)
            done = pageInfo.get('hasNextPage', False) == False
            if not done:
                after = pageInfo['endCursor']
            paydays.extend(this_batch)
            
        return paydays

    # def ql_request_paydays(self, before: str = None, after: str = None):
    #     row_count = (dt.datetime.now() - dt.datetime(2022,6,24,8,0,0)).days + 1
    #     query = "query {"
    #     query += f'paydayStatus {{ nextPaydayTime '
        
    #     if before == 'first':
    #         accounts_str  =  f'paydaySummaries (first: {self.nodes_request_limit}) '
        
    #     elif before != 'null':
    #         accounts_str  =  f'paydaySummaries (last: {self.nodes_request_limit}, before: "{before}") '
    #     elif 'last' in after:
    #         accounts_str  =  f'paydaySummaries ({after}) ' 
        
    #     elif after != 'null':
    #         accounts_str  =  f'paydaySummaries (first: {self.nodes_request_limit}, after: "{after}") ' 
    #     else:
    #         accounts_str = f'paydaySummaries (first: {self.nodes_request_limit})'
    #     query += accounts_str + '{'
    #     query += self.pageInfo()
    #     query += """
    #                 nodes {
    #                     block {
    #                         blockHeight
    #                         blockSlotTime
    #                         blockHash
    #                     }
    #                 }
    #     """
    #     query += ' } } }'

    #     try:
    #         r = requests.post(self.graphql_url, json={'query': query})
    #         if r.status_code == 200:
    #             return  r.json()['data']['paydayStatus']['paydaySummaries']['nodes'], \
    #                     r.json()['data']['paydayStatus']['paydaySummaries']['pageInfo']
                        
                    
    #         else:
    #             print (r.text)

    #     except Exception as e:
    #         print (query, e)
    #         return None


import pymongo
from pymongo import ReplaceOne
from dateutil.parser import isoparse
from ..transaction import Transaction
from ccdexplorer_fundamentals.block import ConcordiumBlockInfo, ConcordiumBlock
from ccdexplorer_fundamentals.transaction import ClassificationResult
import datetime as dt


class Mixin:
    # def store_block_in_mongodb(self, block: ConcordiumBlock, client, version=1):
    #     if not isinstance(block.blockInfo.blockArriveTime, dt.datetime):
    #         block.blockInfo.blockArriveTime    = isoparse(block.blockInfo.blockArriveTime)
    #     if not isinstance(block.blockInfo.blockReceiveTime, dt.datetime):
    #         block.blockInfo.blockReceiveTime   = isoparse(block.blockInfo.blockReceiveTime)
    #     if not isinstance(block.blockInfo.blockSlotTime, dt.datetime):
    #         block.blockInfo.blockSlotTime      = isoparse(block.blockInfo.blockSlotTime)

    #     summary = block.blockSummary
    #     try:
    #         summary.updates['chainParameters']['microGTUPerEuro']['numerator'] = \
    #         str(summary.updates['chainParameters']['microGTUPerEuro']['numerator'])
    #     except:
    #         pass

    #     for k, v in summary.updates.items():
    #         try:
    #             summary[k] = isoparse(v)
    #         except Exception:
    #             pass

    #     for item in summary.specialEvents:
    #         for k, v in item.items():
    #             try:
    #                 summary[k] = isoparse(v)
    #             except Exception:
    #                 pass

    #     if version > 1:
    #         del summary.updates['keys']

    #     summary_to_store = summary.__dict__
    #     if summary_to_store['finalizationData']:
    #         summary_to_store['finalizationData'] = summary_to_store['finalizationData'].__dict__

    #     _id = block.blockInfo.blockHash
    #     mongo_block = {
    #         '_id': _id,
    #         'blockHeight': block.blockInfo.blockHeight,
    #         'blockInfo': block.blockInfo.__dict__,
    #         'blockSummary': summary_to_store}
    #     try:
    #         query = {"_id": _id }
    #         self.collection_blocks.replace_one( query, mongo_block, upsert=True)
    #     except Exception as e:
    #         pass

    #     transactions = summary.transactionSummaries
    #     for tx in transactions:
    #         _id = tx['hash']
    #         tx['blockInfo'] = block.blockInfo.__dict__
    #         tx['_id'] = _id
    #         try:
    #             query = {"_id": _id }
    #             self.collection_transactions.replace_one(query, tx, upsert=True)
    #         except Exception as e:
    #             pass

    #     # finally, store account_tx_link if needed
    #     self.fill_accounts_involved(transactions, client, version)

    def store_blockInfo_in_mongodb(self, block: ConcordiumBlock, client, version=1):
        if version == 2:
            # if not isinstance(block.blockInfo.blockArriveTime, dt.datetime):
            #     block.blockInfo.blockArriveTime    = isoparse(block.blockInfo.blockArriveTime)
            # if not isinstance(block.blockInfo.blockReceiveTime, dt.datetime):
            #     block.blockInfo.blockReceiveTime   = isoparse(block.blockInfo.blockReceiveTime)
            if not isinstance(block.blockInfo.blockSlotTime, dt.datetime):
                block.blockInfo.blockSlotTime = isoparse(block.blockInfo.blockSlotTime)

        summary = block.blockSummary

        _id = block.blockInfo.blockHash
        tx_hashes = [x["hash"] for x in summary.transactionSummaries]
        mongo_block = block.blockInfo.__dict__
        mongo_block["transactionHashes"] = tx_hashes
        try:
            query = {"_id": _id}
            self.collection_blocks.replace_one(query, mongo_block, upsert=True)
        except Exception as e:
            pass

        transactions = summary.transactionSummaries
        for tx in transactions:
            _id = tx["hash"]
            if version == 2:
                tx["blockHeight"] = block.blockInfo.blockHeight
                tx["blockSlotTime"] = block.blockInfo.blockSlotTime
            tx["_id"] = _id
            try:
                query = {"_id": _id}
                self.collection_transactions.replace_one(query, tx, upsert=True)
            except Exception as e:
                pass

        # finally, generate_indices_based_on_transactions
        self.generate_indices_based_on_transactions(transactions, client, version)

    # def possibly_update_tx_with_new_blockInfo_in_mongodb(self, block: ConcordiumBlock, tx, client, version=1):
    #     if not isinstance(block.blockInfo.blockArriveTime, dt.datetime):
    #         block.blockInfo.blockArriveTime    = isoparse(block.blockInfo.blockArriveTime)
    #     if not isinstance(block.blockInfo.blockReceiveTime, dt.datetime):
    #         block.blockInfo.blockReceiveTime   = isoparse(block.blockInfo.blockReceiveTime)
    #     if not isinstance(block.blockInfo.blockSlotTime, dt.datetime):
    #         block.blockInfo.blockSlotTime      = isoparse(block.blockInfo.blockSlotTime)

    #     summary = block.blockSummary

    #     _id = block.blockInfo.blockHash

    #     transactions = summary.transactionSummaries
    #     for tx in transactions:
    #         _id = tx['hash']
    #         tx['blockInfo'] = block.blockInfo.__dict__
    #         tx['_id'] = _id
    #         try:
    #             query = {"_id": _id }
    #             self.collection_transactions.replace_one(query, tx, upsert=True)
    #         except Exception as e:
    #             pass

    #     # finally, store account_tx_link if needed
    #     self.fill_accounts_involved(transactions, client, version)

    def generate_indices_based_on_transactions(self, transactions, client, version):
        """
        Given a list of transactions, apply rules to determine which index needs to be updated.
        Add this to a to_be_sent_to_mongo list and do insert_many.
        """
        list_accounts_involved_all = []
        list_accounts_involved_transfer = []
        list_contracts_involved = []
        # print (f"Generating indices for {len(transactions):,.0f} transactions...")
        for tx in transactions:
            classificationResult, classified_tx = (
                Transaction(client).init_from_node(tx).classify_transaction_for_bot()
            )

            if version == 1:
                pass
            elif version == 2:
                index_all, index_transfer = self.determine_account_indices(
                    classified_tx, tx, classificationResult, version
                )
                list_accounts_involved_all.append(
                    ReplaceOne({"_id": index_all["_id"]}, index_all, upsert=True)
                )
                if index_transfer:
                    list_accounts_involved_transfer.append(
                        ReplaceOne(
                            {"_id": index_transfer["_id"]}, index_transfer, upsert=True
                        )
                    )

                if classificationResult.contracts_involved:
                    for contract in classificationResult.list_of_contracts_involved:
                        index_contract = self.determine_contract_indices(
                            classified_tx, tx, classificationResult, contract, version
                        )
                        list_contracts_involved.append(
                            ReplaceOne(
                                {"_id": index_contract["_id"]},
                                index_contract,
                                upsert=True,
                            )
                        )

        try:
            if len(list_accounts_involved_all) > 0:
                result = self.collection_involved_accounts_all.bulk_write(
                    list_accounts_involved_all
                )
                print(
                    f"B: {tx['blockHeight']:,.0f} | Match {result.matched_count:,.0f} | Mod {result.modified_count:,.0f} | Upsert {result.upserted_count:2,.0f} in _all."
                )
        except Exception as e:
            print(e)

        try:
            if len(list_accounts_involved_transfer) > 0:
                result = self.collection_involved_accounts_transfer.bulk_write(
                    list_accounts_involved_transfer
                )
                print(
                    f"B: {tx['blockHeight']:,.0f} | Match {result.matched_count:,.0f} | Mod {result.modified_count:,.0f} | Upsert {result.upserted_count:2,.0f} in _transfer."
                )
        except Exception as e:
            print(e)

        try:
            if len(list_contracts_involved) > 0:
                result = self.collection_involved_contracts.bulk_write(
                    list_contracts_involved
                )
                print(
                    f"B: {tx['blockHeight']:,.0f} | Match {result.matched_count:,.0f} | Mod {result.modified_count:,.0f} | Upsert {result.upserted_count:2,.0f} in _contracts."
                )
        except Exception as e:
            print(e)

    # def fill_accounts_involved(self, transactions, client, version=1):
    #     for tx in transactions:
    #     #     tx_already_done = self.collection_accounts_involved_all.find_one({"_id": tx['blockInfo']['blockHash']})

    #     #     if not tx_already_done:
    #         classificationResult, classified_tx = Transaction(client).init_from_node(tx).classify_transaction_for_bot()

    #         if version == 1:
    #             if classificationResult.sender and classificationResult.receiver:
    #                 self.store_account_tx_link(classified_tx, classificationResult, version)

    #         elif version == 2:
    #             self.determine_account_indices(classified_tx, classificationResult, version)

    #             if classificationResult.contracts_involved:
    #                 for contract in classificationResult.list_of_contracts_involved:
    #                     self.store_contract_tx_link_all(classified_tx, classificationResult, contract, version)

    def determine_account_indices(
        self,
        classified_tx: Transaction,
        tx,
        classificationResult: ClassificationResult,
        version=1,
    ):
        _id = classificationResult.txHash
        dct = {
            "_id": _id,
            "sender": classificationResult.sender,
            "receiver": classificationResult.receiver,
            "sender_canonical": (
                classificationResult.sender[:29]
                if classificationResult.sender
                else None
            ),
            "receiver_canonical": (
                classificationResult.receiver[:29]
                if classificationResult.receiver
                else None
            ),
            "amount": classificationResult.amount,
            "type": classificationResult.type,
            "contents": classificationResult.contents,
            "blockHeight": tx["blockHeight"],
        }
        if version > 1:
            memo = None
            if "events" in classified_tx.result:
                for event in classified_tx.result["events"]:
                    if "memo" in event:
                        memo = event["memo"]

            if memo:
                dct.update({"memo": memo})
        if classificationResult.accounts_involved_transfer:
            return dct, dct
        else:
            return dct, None

    def determine_contract_indices(
        self,
        classified_tx: Transaction,
        tx,
        classificationResult: ClassificationResult,
        contract,
        version=1,
    ):
        _id = f"{classificationResult.txHash}-<{contract[0]},{contract[1]}>"
        dct = {
            "_id": _id,
            "index": contract[0],
            "subindex": contract[1],
            "contract": f"<{contract[0]},{contract[1]}>",
            "type": classificationResult.type,
            "contents": classificationResult.contents,
            "blockHeight": tx["blockHeight"],
        }

        return dct

    # def store_account_tx_link(self,
    #     classified_tx: Transaction,
    #     classificationResult: ClassificationResult,
    #     version=1):
    #     _id = classificationResult.txHash
    #     dct = {

    #         "_id":      _id,
    #         "sender":   classificationResult.sender,
    #         "receiver": classificationResult.receiver,
    #         "amount":   classificationResult.amount,
    #         "type":     classificationResult.type,
    #         "contents": classificationResult.contents,
    #         "blockHeight": classified_tx.block['blockHeight']
    #     }
    #     if version > 1:
    #         memo = None
    #         if 'events' in classified_tx.result:
    #             for event in classified_tx.result['events']:
    #                 if 'memo' in event:
    #                     memo = event['memo']

    #         if memo:
    #             dct.update(
    #                 {
    #                     'memo': memo
    #                 }
    #             )
    #     try:
    #         if version == 1:
    #             self.collection_accounts_involved.insert_one(dct)
    #         elif version == 2:
    #             if classificationResult.accounts_involved_transfer:
    #                 query = {"_id": _id }
    #                 self.collection_accounts_involved_transfer.replace_one(query, dct, upsert=True)

    #             if classificationResult.accounts_involved_all:
    #                 query = {"_id": _id }
    #                 self.collection_accounts_involved_all.replace_one(query, dct, upsert=True)
    #     except Exception as e:
    #         print (e)

    # def store_contract_tx_link_all(self,
    #     classified_tx: Transaction,
    #     classificationResult: ClassificationResult,
    #     contract,
    #     version=1):
    #     _id =  f"{classificationResult.txHash}-<{contract[0]},{contract[1]}>"
    #     dct = {

    #         "_id":      _id,
    #         "index":    contract[0],
    #         "subindex": contract[1],
    #         "contract": f"<{contract[0]},{contract[1]}>",
    #         "type":     classificationResult.type,
    #         "contents": classificationResult.contents,
    #         "blockHeight": classified_tx.block['blockHeight']
    #     }

    #     try:
    #         if version == 1:
    #             pass

    #         elif version == 2:
    #             query = {"_id": _id }
    #             self.collection_contracts_involved.replace_one(query, dct, upsert=True)
    #     except Exception as e:
    #         print (e)

    def search_tx_higher_than_block_height(self, block_height_start, block_height_end):
        pipeline = [
            {"$match": {"blockInfo.blockHeight": {"$gt": block_height_start}}},
            {"$match": {"blockInfo.blockHeight": {"$lt": block_height_end}}},
            # {'$project': {
            #     '_id': 1
            #     }
            # }
        ]

        return pipeline



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: health.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0chealth.proto\x12\x11\x63oncordium.health\"\x13\n\x11NodeHealthRequest\"\x14\n\x12NodeHealthResponse2^\n\x06Health\x12T\n\x05\x43heck\x12$.concordium.health.NodeHealthRequest\x1a%.concordium.health.NodeHealthResponseB/\n\x16\x63om.concordium.grpc.v2P\x01\xaa\x02\x12\x43oncordium.Grpc.V2b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'health_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.concordium.grpc.v2P\001\252\002\022Concordium.Grpc.V2'
  _globals['_NODEHEALTHREQUEST']._serialized_start=35
  _globals['_NODEHEALTHREQUEST']._serialized_end=54
  _globals['_NODEHEALTHRESPONSE']._serialized_start=56
  _globals['_NODEHEALTHRESPONSE']._serialized_end=76
  _globals['_HEALTH']._serialized_start=78
  _globals['_HEALTH']._serialized_end=172
# @@protoc_insertion_point(module_scope)
